/* edit-shapebg-row.c
 *
 * Copyright 2023 Jonathan Blandford
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "crosswords-config.h"
#include <libipuz/libipuz.h>
#include <glib/gi18n-lib.h>
#include <adwaita.h>
#include "edit-shapebg-row.h"
#include "xword-state.h"


enum
  {
    SHAPEBG_CHANGED,
    N_SIGNALS
  };

static guint obj_signals[N_SIGNALS] = { 0 };


struct _EditShapebgRow
{
  AdwActionRow parent_instance;

  GtkWidget *shapebg_popover;
  GtkWidget *popover_grid;
  GArray *shapebg_buttons;
};


G_DEFINE_FINAL_TYPE (EditShapebgRow, edit_shapebg_row, ADW_TYPE_ACTION_ROW);


static void edit_shapebg_row_init       (EditShapebgRow      *self);
static void edit_shapebg_row_class_init (EditShapebgRowClass *klass);
static void edit_shapebg_row_activate   (AdwActionRow        *action_row);


typedef struct _ShapebgMapping
{
  const gchar *shapebg;
  guint column;
  guint row;
} ShapebgMapping;

static ShapebgMapping shapebg_mapping[] = {
  { "circle", 0, 0 },
  { "star", 1, 0 },
  { "square", 2, 0 },
  { "rhombus", 3, 0 },
  { "club", 0, 1 },
  { "diamond", 1, 1 },
  { "heart", 2, 1 },
  { "spade", 3, 1 },
  { "arrow-left", 0, 2 },
  { "arrow-right", 1, 2 },
  { "arrow-up", 2, 2 },
  { "arrow-down", 3, 2 },
  { "triangle-left", 0, 3 },
  { "triangle-right", 1, 3 },
  { "triangle-up", 2, 3 },
  { "triangle-down", 3, 3 },
  { "slash", 0, 4 },
  { "backslash", 1, 4 },
  { "x", 2, 4 },
  { NULL, 0, 0 }
};


static void
set_up_shapebg_button (EditShapebgRow *self,
                      ShapebgMapping   shapebg_mapping)
{
  GtkWidget *button;
  GtkWidget *image;
  g_autofree const gchar *resource_path = NULL;

  resource_path = g_strdup_printf ("/org/gnome/Crosswords/icons/shapebg/%s.svg",
                                   shapebg_mapping.shapebg);
  button = gtk_button_new ();
  image = gtk_image_new_from_resource (resource_path);
  gtk_button_set_child (GTK_BUTTON (button), image);
  gtk_grid_attach (GTK_GRID (self->popover_grid),
                   button,
                   shapebg_mapping.column, shapebg_mapping.row,
                   1, 1);
}

static void
edit_shapebg_row_init (EditShapebgRow *self)
{
  GtkWidget *child;
  gtk_widget_init_template (GTK_WIDGET (self));

  for (guint i = 0; shapebg_mapping[i].shapebg ; i ++)
    set_up_shapebg_button (self, shapebg_mapping[i]);

  /* FIXME(hack): This is a pretty tough hack to get the title_box of
   * the ActionRow to expand horizontally. It's a private widget, as
   * is adw_action_row_set_expand_suffixes() which is how libadwaita
   * handles it for its widgets. */
  child = gtk_widget_get_first_child (GTK_WIDGET (self));
  if (!GTK_IS_BOX (child) ||
      g_strcmp0 ("header", gtk_buildable_get_buildable_id (GTK_BUILDABLE (child))))
    {
      g_warning ("AdwActionRow's internal structure has changed");
      return;
    }

  gtk_widget_set_hexpand (child, TRUE);
}
                                         
static void
edit_shapebg_row_class_init (EditShapebgRowClass *klass)
{
  GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (klass);
  AdwActionRowClass *action_row_class = ADW_ACTION_ROW_CLASS (klass);

  action_row_class->activate = edit_shapebg_row_activate;

  gtk_widget_class_set_template_from_resource (widget_class, "/org/gnome/Crosswords/edit-shapebg-row.ui");

  gtk_widget_class_bind_template_child (widget_class, EditShapebgRow, shapebg_popover);
  gtk_widget_class_bind_template_child (widget_class, EditShapebgRow, popover_grid);

  obj_signals[SHAPEBG_CHANGED] =
    g_signal_new ("shapebg_row-type-changed",
                  EDIT_TYPE_SHAPEBG_ROW,
                  G_SIGNAL_RUN_FIRST,
                  0,
                  NULL, NULL,
                  NULL,
                  G_TYPE_NONE, 1,
                  G_TYPE_STRING); /* shapebg */
  gtk_widget_class_set_layout_manager_type (widget_class, GTK_TYPE_BOX_LAYOUT);

  gtk_widget_class_set_css_name (widget_class, "row");
}

static void
edit_shapebg_row_activate (AdwActionRow *action_row)
{
  EditShapebgRow *self;

  self = EDIT_SHAPEBG_ROW (action_row);

  gtk_popover_popup (GTK_POPOVER (self->shapebg_popover));
}
                                


/* Public methods */

void
edit_shapebg_row_update (EditShapebgRow *self,
                         XwordState     *grid_state)
{
  g_assert (grid_state->xword != NULL);
}
