/* xword-state.h
 *
 * Copyright 2022 Jonathan Blandford <jrb@gnome.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 */


#pragma once


#include <gtk/gtk.h>
#include <libipuz/libipuz.h>

#include "cell-array.h"
#include "crosswords-quirks.h"

G_BEGIN_DECLS

#define XWORD_STATE_BLOCK_CHARS ".#"
#define XWORD_STATE_SKIP_CHARS  "?"


#define XWORD_STATE_HAS_CURSOR(state)       (state->mode==XWORD_STATE_EDIT||state->clue.direction!=IPUZ_CLUE_DIRECTION_NONE)
#define NONE_HAS_FOCUS(state)               ((state != NULL) && \
		                             (state->quirks != NULL) && \
		                             (crosswords_quirks_get_focus_location(state->quirks)==QUIRKS_FOCUS_LOCATION_NONE))
#define ACROSTIC_MAIN_GRID_HAS_FOCUS(state) ((state != NULL) && \
		                             (IPUZ_IS_ACROSTIC (state->xword)) && \
		                             (state->quirks != NULL) && \
					     ((crosswords_quirks_get_focus_location(state->quirks)==QUIRKS_FOCUS_LOCATION_MAIN_GRID)))
#define ACROSTIC_CLUE_GRID_HAS_FOCUS(state) ((state != NULL) && \
		                             (IPUZ_IS_ACROSTIC (state->xword)) && \
		                             (state->quirks != NULL) && \
		                             (crosswords_quirks_get_focus_location(state->quirks)==QUIRKS_FOCUS_LOCATION_CLUE_GRID))
#define GET_QUOTE_CLUE(xword)               (IPUZ_IS_ACROSTIC(xword)?ipuz_acrostic_get_quote_clue(IPUZ_ACROSTIC(xword)):NULL)
#define XWORD_KIND_DESTRUCTIVE(kind)        (kind==XWORD_CMD_KIND_DELETE||kind==XWORD_CMD_KIND_BACKSPACE)

typedef enum
{
  XWORD_CMD_KIND_NONE,
  XWORD_CMD_KIND_UP,
  XWORD_CMD_KIND_DOWN,
  XWORD_CMD_KIND_LEFT,
  XWORD_CMD_KIND_RIGHT,
  XWORD_CMD_KIND_FORWARD,
  XWORD_CMD_KIND_BACK,
  XWORD_CMD_KIND_FORWARD_CLUE,
  XWORD_CMD_KIND_BACK_CLUE,
  XWORD_CMD_KIND_BACKSPACE,
  XWORD_CMD_KIND_DELETE,
  XWORD_CMD_KIND_SWITCH,
} XwordCmdKind;

typedef enum
{
  XWORD_STATE_SOLVE,       /* @ Solve a crossword @ */
  XWORD_STATE_BROWSE,      /* @ Browse through a board without modifying it @ */
  XWORD_STATE_EDIT,        /* @ Edit the grid @ */
  XWORD_STATE_EDIT_BROWSE, /* @ Browse through an editable board without modifying it @ */
  XWORD_STATE_SELECT,      /* @ Select cells @ */
  XWORD_STATE_VIEW,        /* @ Display a board with no interaction @ */
} XwordStateMode;


typedef enum
{
  XWORD_REVEAL_NONE,         /* @ Reveal nothing. Display the crossword as normal @ */
  XWORD_REVEAL_ERRORS_BOARD, /* @ Reveal all incorrect guesses on the board @ */
  XWORD_REVEAL_ERRORS_CLUE,  /* @ Reveal all incorrect guesses in the current clue @ */
  XWORD_REVEAL_ERRORS_CELL,  /* @ Reveal all incorrect guesses in the current cell @ */
  XWORD_REVEAL_ALL,          /* @ Reveal all unsolved letters. Effectively give up @ */
} XwordRevealMode;

typedef enum
{
  XWORD_SELECTION_SELECT,    /* @ Select a cell @ */
  XWORD_SELECTION_TOGGLE,    /* @ Toggle selection @ */
  XWORD_SELECTION_EXTEND,    /* @ Extend the selection @ */
} XwordSelectionMode;

typedef struct
{
  XwordStateMode mode;
  IPuzCrossword *xword;

  /* Current keyboard state */
  IPuzCellCoord cursor;
  IPuzClueId clue; /* Current selected direction and clue ID, or NONE */

  /* Modifiers to the current display and behavior */
  XwordRevealMode reveal_mode;
  gboolean skip_filled;
  CellArray *saved_selection;
  CellArray *selected_cells;
  CrosswordsQuirks *quirks;
} XwordState;


XwordState *xword_state_new                (IPuzCrossword      *xword,
                                            CrosswordsQuirks   *quirks,
                                            XwordStateMode      mode) G_GNUC_WARN_UNUSED_RESULT;
void        xword_state_free               (XwordState         *state);
XwordState *xword_state_dehydrate          (XwordState         *state);
XwordState *xword_state_hydrate            (XwordState         *state,
                                            IPuzCrossword      *xword,
                                            CrosswordsQuirks   *quirk);
XwordState *xword_state_swap_xword         (XwordState         *state,
                                            IPuzCrossword      *xword);
XwordState *xword_state_replace            (XwordState         *old_state,
                                            XwordState         *new_state) G_GNUC_WARN_UNUSED_RESULT;
                                                               
XwordState *xword_state_change_mode        (XwordState         *state,
                                            XwordStateMode      mode) G_GNUC_WARN_UNUSED_RESULT;
XwordState *xword_state_clue_selected      (XwordState         *state,
                                            IPuzClue           *clue) G_GNUC_WARN_UNUSED_RESULT;
XwordState *xword_state_cell_selected      (XwordState         *state,
                                            IPuzCellCoord       coord) G_GNUC_WARN_UNUSED_RESULT;
XwordState *xword_state_do_command         (XwordState         *state,
                                            XwordCmdKind        kind) G_GNUC_WARN_UNUSED_RESULT;
XwordState *xword_state_guess              (XwordState         *state,
                                            const gchar        *guess) G_GNUC_WARN_UNUSED_RESULT;
XwordState *xword_state_guess_at_cell      (XwordState         *state,
                                            const gchar        *guess,
                                            IPuzCellCoord       coord) G_GNUC_WARN_UNUSED_RESULT;
XwordState *xword_state_guess_word         (XwordState         *state,
                                            IPuzCellCoord       coord,
                                            IPuzClueDirection   direction,
                                            const gchar        *word) G_GNUC_WARN_UNUSED_RESULT;
XwordState *xword_state_set_cell_type      (XwordState         *state,
                                            IPuzCellCoord       coord,
                                            IPuzCellCellType    cell_type) G_GNUC_WARN_UNUSED_RESULT;
XwordState *xword_state_select_cell        (XwordState         *state,
                                            IPuzCellCoord       coord) G_GNUC_WARN_UNUSED_RESULT;
XwordState *xword_state_select_toggle      (XwordState         *state,
                                            IPuzCellCoord       coord) G_GNUC_WARN_UNUSED_RESULT;
XwordState *xword_state_select_drag_start  (XwordState         *state,
                                            IPuzCellCoord       anchor_coord,
                                            IPuzCellCoord       new_coord,
                                            XwordSelectionMode  mode) G_GNUC_WARN_UNUSED_RESULT;
XwordState *xword_state_select_drag_update (XwordState         *state,
                                            IPuzCellCoord       anchor_coord,
                                            IPuzCellCoord       new_coord,
                                            XwordSelectionMode  mode) G_GNUC_WARN_UNUSED_RESULT;
XwordState *xword_state_select_drag_end    (XwordState         *state) G_GNUC_WARN_UNUSED_RESULT;
XwordState *xword_state_select_all         (XwordState         *state) G_GNUC_WARN_UNUSED_RESULT;
XwordState *xword_state_select_none        (XwordState         *state) G_GNUC_WARN_UNUSED_RESULT;
XwordState *xword_state_select_invert      (XwordState         *state) G_GNUC_WARN_UNUSED_RESULT;
XwordState *xword_state_set_reveal_mode    (XwordState         *state,
                                            XwordRevealMode     reveal_mode) G_GNUC_WARN_UNUSED_RESULT;
XwordState *xword_state_set_bars           (XwordState         *state,
                                            IPuzCellCoord       coord,
                                            IPuzStyleSides      bars,
                                            gboolean            update_cursor);
                                                               
void        xword_state_assert_equal       (const XwordState   *a,
                                            const XwordState   *b);
                                                               
void        xword_state_print              (const XwordState   *state,
                                            gboolean            print_members);


G_END_DECLS
